/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CVFDCAMERASPEC_H_
#define _CVFDCAMERASPEC_H_

#include "Object.h"
#include "ObjectParser.h"
#include "misc.h"
#include <cmath>

using namespace FD;

class CvfdCameraSpec : public Object{
  
 public:
  int   width;
  int   height;
  float deg[2];
  float offset[2];
  float offset_scale[2];
  float scale_r;
  float pos[3];

  CvfdCameraSpec(){
  }

  CvfdCameraSpec(const CvfdCameraSpec& input) {
    width = input.width;
    height = input.height;
    for(int cnt = 0; cnt < 2; cnt++) deg[cnt] = input.deg[cnt];
    for(int cnt = 0; cnt < 2; cnt++) offset[cnt] = input.offset[cnt];
    for(int cnt = 0; cnt < 2; cnt++) offset_scale[cnt] = input.offset[cnt];
    scale_r = input.scale_r;
    for(int cnt = 0; cnt < 3; cnt++) pos[cnt] = input.pos[cnt];
    SetPosition(input.pos[0], input.pos[1], input.pos[2]);
  }
      
  ~CvfdCameraSpec(){
  }

  ///////////////////////////////////////
  //
  // FlowDesigner related member functions
		
  /**Serialize (binary) the object to a stream*/
  virtual void serialize(std::ostream &out) const {
  }
  
  /**Unserialize (binary) the object from a stream*/
  virtual void unserialize(std::istream &in){
  }
		
  /**Generic print function*/
  virtual void printOn(std::ostream &out=std::cout) const {
  }
     
  /**Generic read function*/
  virtual void readFrom(std::istream &in=std::cin){
  }

  ///////////////////////////////////////
  //
  // Image mapping related member functions

  float deg2rad(float x) const { return ((x)*M_PI)/180.0;}

  float rad2deg(float x) const { return ((x)*180.0)/M_PI;}

  int myround(float x) const {return (int)floor(x + 0.5);}

  float pt2rad_x(int x) const
  {
    return atan2(x-(float)(width-1)/2.0, (float)(width-1)/2.0/tan(deg2rad(deg[0]/2.0)));
  }
  float pt2rad_y(int y) const
  {
    return atan2(y-(float)(height-1)/2.0,(float)(height-1)/2.0/tan(deg2rad(deg[1]/2.0)));
  }
  int rad2pt_x(float rad_x) const
  {
    return myround((float)(width-1)/2.0 +(float)(width-1)/2.0*tan(rad_x)/tan(deg2rad(deg[0]/2.0)));
  }
  int rad2pt_y(float rad_y) const
  {
    return myround((float)(height-1)/2.0+(float)(height-1)/2.0*tan(rad_y)/tan(deg2rad(deg[1]/2.0)));
  }

  void mapping(const float pt[3], int p1[2]) const
  {
    p1[0] = rad2pt_x(atan2(pt[0],pt[2])) - width / 2;
    p1[1] = rad2pt_y(atan2(pt[1],pt[2])) - height / 2;
    float rad = atan2(p1[1],p1[0]);
    float r = sqrt(pow(p1[0],2) + pow(p1[1],2));
    float r2 = r * scale_r;
    p1[0] = myround(r2 * cos(rad)) + width / 2;
    p1[0] += (int)((float)p1[0] - offset[0]) / offset_scale[0];
    p1[1] = myround(r2 * sin(rad)) + height / 2;
    p1[1] += (int)((float)p1[1] - offset[1]) / offset_scale[1];
    return;
  }

  ///////////////////////////////////////
  //
  // Transform matrix related member functions

  void printMatrix(const float a[4][4]) const
  {
    for(int row=0;row<4;row++){
      for(int col=0;col<4;col++){
	printf("%.06f ",a[row][col]);
      }
      printf("\n");
    }
  }

  void multiply(const float a[4][4],const float p0[4],float p1[4]) const
  {
    for(int row=0;row<4;row++){
      p1[row] = 0;
      for(int col=0;col<4;col++){
	p1[row] += a[row][col]*p0[col];
      }
    }
    return;
  }
      
  void translate(const float pt[3], float p2[3]) const
  {
    float p0[4];
    *(p0+0) = *(pt+0);
    *(p0+1) = *(pt+1);
    *(p0+2) = *(pt+2);
    *(p0+3) = 1.0;	
    float p1[4];
    multiply(m_a,p0,p1);
    *(p2+0) = *(p1+0);
    *(p2+1) = *(p1+1);
    *(p2+2) = *(p1+2);
    return;
  }

  void inverse(const float pt[3], float p2[3]) const
  {
    float p0[4];
    *(p0+0) = *(pt+0);
    *(p0+1) = *(pt+1);
    *(p0+2) = *(pt+2);
    *(p0+3) = 1.0;	
    float p1[4];
    multiply(m_ia,p0,p1);
    *(p2+0) = *(p1+0);
    *(p2+1) = *(p1+1);
    *(p2+2) = *(p1+2);
    return;
  }

  void printMatrixA() const
  {
    printf("m_a ---\n");
    printMatrix(m_a);
  }
  void printMatrixIA() const
  {
    printf("m_ia ---\n");
    printMatrix(m_ia);
  }
  
  void SetPosition(const CvfdCameraSpec& input)
  {
    SetPosition(input.pos[0], input.pos[1], input.pos[2]);
  }
  
  void SetPosition(float x, float y, float z)
  {
    int row=0;
    int col=0;
    
    m_a[row][col++] = 1;
    m_a[row][col++] = 0;
    m_a[row][col++] = 0;
    m_a[row][col++] = x;
    row++;col=0;
    
    m_a[row][col++] = 0;
    m_a[row][col++] = 1;
    m_a[row][col++] = 0;
    m_a[row][col++] = y;
    row++;col=0;
    
    m_a[row][col++] = 0;
    m_a[row][col++] = 0;
    m_a[row][col++] = 1;
    m_a[row][col++] = z;
    row++;col=0;
    
    m_a[row][col++] = 0;
    m_a[row][col++] = 0;
    m_a[row][col++] = 0;
    m_a[row][col++] = 1;
    row++;col=0;
    
    row=0;col=0;
    
    m_ia[row][col++] = 1;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = -x;
    row++;col=0;
    
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 1;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = -y;
    row++;col=0;

    m_ia[row][col++] = 0;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 1;
    m_ia[row][col++] = -z;
    row++;col=0;
    
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 0;
    m_ia[row][col++] = 1;
    row++;col=0;
  }

  ///////////////////////////////////////
  //
  // HUE related member functions

  float computeChannel( float fChannelVal, float fLowVal, float fHighVal )
  {
    if (6.0 * fChannelVal < 1.0) return fLowVal + (fHighVal - fLowVal) * 6.0 * fChannelVal;
    if (2.0 * fChannelVal < 1.0) return fHighVal;
    if (3.0 * fChannelVal < 2.0) return fLowVal + (fHighVal - fLowVal) * ((2.0 / 3.0) - fChannelVal) * 6.0;
    return fLowVal;
  }

  void convertHslToRgb( int nH, int nS, int nL, int* pR, int* pG, int* pB )
  {
    float fR, fG, fB, fLowValue, fHighValue, fH, fS, fL;
    
    //Scale each value to 0.0 - 1.0, from type int:
    fH = nH / 360.0;  //Hue is represented as a range of 360 degrees
    fS = nS / 256.0;  //Saturation 
    fL = nL / 256.0;  //Lightness
    
    if (fS == 0)         //Saturation of 0 means a shade of grey
      {
	fR  = fG = fB = fL;
      }
    else
      {
	if (fL < 0.5) fHighValue = fL * (1.0 + fS);
	else fHighValue = (fL + fS) - (fL * fS);
	
	fLowValue = 2.0 * fL - fHighValue;
	
	fR = fH + 1.0 / 3.0;
	if (fR > 1.0) fR-= 1.0;
	
	fG = fH;
	
	fB = fH - 1.0 / 3.0;
	if (fB < 0) fB += 1.0;
	
	fR = computeChannel(fR, fLowValue, fHighValue);
	fG = computeChannel(fG, fLowValue, fHighValue);
	fB = computeChannel(fB, fLowValue, fHighValue);
      }
    
    *pR = int(fR * 256.0);
    *pG = int(fG * 256.0);
    *pB = int(fB * 256.0);
  }
  
 protected:
  float m_a[4][4];
  float m_ia[4][4];
      
};

#endif
