/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

using namespace std;
using namespace FD;

class CvfdhCalcOpticalFlow;
   
DECLARE_NODE(CvfdhCalcOpticalFlow)
/*Node
 *
 * @name CvfdhCalcOpticalFlow
 * @category FDOPENCV_HARK:Feature
 * @description CvfdhCalcOpticalFlow
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 * @parameter_name OPTICAL_FLOW_TYPE
 * @parameter_type string
 * @parameter_value PyrLK
 * @parameter_list BM:HS:LK:PyrLK
 * @parameter_description Optical flow type
 *
 * @parameter_name BM_BLOCK_SIZE
 * @parameter_type int
 * @parameter_value 200
 * @parameter_valid OPTICAL_FLOW_TYPE=BM
 * @parameter_description Block size
 *
 * @parameter_name BM_SHIFT_SIZE
 * @parameter_type int
 * @parameter_value 100
 * @parameter_valid OPTICAL_FLOW_TYPE=BM
 * @parameter_description Shift size
 *
 * @parameter_name BM_MAX_SIZE
 * @parameter_type int
 * @parameter_value 400
 * @parameter_valid OPTICAL_FLOW_TYPE=BM
 * @parameter_description Max size
 *
 END*/
   
   
class CvfdhCalcOpticalFlow : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
  string optical_flow_type;
  int bm_block_size;
  int bm_shift_size;
  int bm_max_size;

public:
  CvfdhCalcOpticalFlow(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcInID = addInput("SRCIN");
    m_featOutID = addOutput("FEATOUT");
    optical_flow_type = object_cast<String>(parameters.get("OPTICAL_FLOW_TYPE"));
    bm_block_size = dereference_cast<int>(parameters.get("BM_BLOCK_SIZE"));
    bm_shift_size = dereference_cast<int>(parameters.get("BM_SHIFT_SIZE"));
    bm_max_size = dereference_cast<int>(parameters.get("BM_MAX_SIZE"));
    inputsCache[m_imageInID].lookAhead = 1;
    inOrder = true;
  }      

  virtual int translateInput(string inputName) {
    if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;


    RCPtr<Map<int, ObjectRef> > image_map0(new Map<int, ObjectRef>);
    RCPtr<Map<int, ObjectRef> > image_map1(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp0;
    IplImage* img_in_tmp1;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr0 = getInput(m_imageInID,count);   
      if(imagePtr0->getImage() == NULL){
	return;
      }
      img_in_tmp0 = cvCloneImage(imagePtr0->getImage());
      CvfdImage* image0= new CvfdImage(img_in_tmp0);  
      (*image_map0)[0] = image0;

      RCPtr<CvfdImage> imagePtr1 = getInput(m_imageInID,count + 1);   
      if(imagePtr1->getImage() == NULL){
	return;
      }
      img_in_tmp1 = cvCloneImage(imagePtr1->getImage());
      CvfdImage* image1= new CvfdImage(img_in_tmp1);  
      (*image_map1)[0] = image1;

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr0 = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > imageMapPtr1 = getInput(m_imageInID, count + 1);
      Map<int, ObjectRef>::iterator it = imageMapPtr0->begin();
      for (; it != imageMapPtr0->end(); ++it) {
        CvfdImage& imagePtr0 = object_cast<CvfdImage>(it->second);
	if ((*imageMapPtr1).find(it->first) == (*imageMapPtr1).end()) continue;
	const CvfdImage& imagePtr1 = 
	  object_cast<CvfdImage>((*imageMapPtr1).find(it->first)->second);

	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;

	img_in_tmp0 = cvCloneImage(imagePtr0.getImage());
	CvfdImage* image0 = new CvfdImage(img_in_tmp0);  
	(*image_map0)[it->first] = image0;

	img_in_tmp1 = cvCloneImage(imagePtr1.getImage());
	CvfdImage* image1 = new CvfdImage(img_in_tmp1);  
	(*image_map1)[it->first] = image1;

      }

    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map0->begin();
    for (; img_it != image_map0->end(); ++img_it) {

      IplImage* cv_image0;      
      CvfdImage& imagePtr0 = object_cast<CvfdImage>(img_it->second);
      cv_image0 = cvCloneImage(imagePtr0.getImage());

      IplImage* cv_image1;
      const CvfdImage& imagePtr1 = 
	object_cast<CvfdImage>((*image_map1).find(img_it->first)->second);
      cv_image1 = cvCloneImage(imagePtr1.getImage());

      // IplImage* cv_image;
    
      // ---------------------------------------------------------------
      // main routine : start

      int dx, dy, rows, cols;
      CvMat *velx, *vely;
      CvSize block = cvSize (bm_block_size, bm_block_size);
      CvSize shift = cvSize (bm_shift_size, bm_shift_size);
      CvSize max_range = cvSize (bm_max_size, bm_max_size);
      IplImage *src_img0, *src_img1, *dst_img; 
      CvTermCriteria criteria;
      char *status;
      int corner_count = 150;
      CvPoint2D32f *corners1, *corners2;
      IplImage *eig_img, *temp_img;
      IplImage *prev_pyramid, *curr_pyramid;
    
      src_img0 = cvCreateImage( cvGetSize(cv_image0),IPL_DEPTH_8U,1);
      cvCvtColor(cv_image0, src_img0, CV_BGR2GRAY);
    
      src_img1 = cvCreateImage( cvGetSize(cv_image1),IPL_DEPTH_8U,1);
      cvCvtColor(cv_image1, src_img1, CV_BGR2GRAY);
    
      dst_img = cvCloneImage (cv_image0);  

      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      if (optical_flow_type == "BM"){
      
	// cols = floor((src_img0->width - block.width) / shift.width); 
	// rows = floor((src_img0->height - block.height) / shift.height); 
	rows = int (ceil (double (src_img0->height) / bm_block_size));
	cols = int (ceil (double (src_img0->width) / bm_block_size));
	velx = cvCreateMat (rows, cols, CV_32FC1);
	vely = cvCreateMat (rows, cols, CV_32FC1);
	cvSetZero (velx);
	cvSetZero (vely);
    
	cvCalcOpticalFlowBM (src_img0, src_img1, block, shift, max_range, 0, velx, vely);
    
	for (int i = 0; i < velx->width; i++) {
	  for (int j = 0; j < vely->height; j++) {
	    // dx = (int) cvGetReal2D (velx, j, i);
	    // dy = (int) cvGetReal2D (vely, j, i);
	    // cvLine (dst_img, cvPoint (i * bm_block_size, j * bm_block_size), cvPoint (i * bm_block_size + dx, j * bm_block_size + dy), CV_RGB (255, 0, 0), 1, CV_AA, 0);
	    (*map_contents).push_back((float)i);
	    (*map_contents).push_back((float)j);
	    (*map_contents).push_back((float)cvGetReal2D (velx, j, i));
	    (*map_contents).push_back((float)cvGetReal2D (vely, j, i));
	  }
	}

      }

      if (optical_flow_type == "HS"){

	cols = src_img0->width;
	rows = src_img0->height;
	velx = cvCreateMat (rows, cols, CV_32FC1);
	vely = cvCreateMat (rows, cols, CV_32FC1);
	cvSetZero (velx);
	cvSetZero (vely);
	criteria = cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 64, 0.01);
      
	cvCalcOpticalFlowHS (src_img0, src_img1, 0, velx, vely, 100.0, criteria);
	for (int i = 0; i < cols; i += 5) {
	  for (int j = 0; j < rows; j += 5) {
	    // dx = (int) cvGetReal2D (velx, j, i);
	    // dy = (int) cvGetReal2D (vely, j, i);
	    // cvLine (dst_img, cvPoint (i, j), cvPoint (i + dx, j + dy), CV_RGB (255, 0, 0), 1, CV_AA, 0);
	    (*map_contents).push_back((float)i);
	    (*map_contents).push_back((float)j);
	    (*map_contents).push_back((float)cvGetReal2D (velx, j, i));
	    (*map_contents).push_back((float)cvGetReal2D (vely, j, i));
	  }
	}

      }
    
      if (optical_flow_type == "LK"){

	cols = src_img0->width;
	rows = src_img0->height;
	velx = cvCreateMat (rows, cols, CV_32FC1);
	vely = cvCreateMat (rows, cols, CV_32FC1);
	cvSetZero (velx);
	cvSetZero (vely);
	criteria = cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 64, 0.01);

	cvCalcOpticalFlowLK (src_img0, src_img1, cvSize (15, 15), velx, vely);
	for (int i = 0; i < cols; i += 5) {
	  for (int j = 0; j < rows; j += 5) {
	    // dx = (int) cvGetReal2D (velx, j, i);
	    // dy = (int) cvGetReal2D (vely, j, i);
	    // cvLine (dst_img, cvPoint (i, j), cvPoint (i + dx, j + dy), CV_RGB (255, 0, 0), 1, CV_AA, 0);
	    (*map_contents).push_back((float)i);
	    (*map_contents).push_back((float)j);
	    (*map_contents).push_back((float)cvGetReal2D (velx, j, i));
	    (*map_contents).push_back((float)cvGetReal2D (vely, j, i));
	  }
	} 

      }

      if (optical_flow_type == "PyrLK"){

	eig_img = cvCreateImage (cvGetSize (src_img0), IPL_DEPTH_32F, 1);
	temp_img = cvCreateImage (cvGetSize (src_img0), IPL_DEPTH_32F, 1);
	corners1 = (CvPoint2D32f *) cvAlloc (corner_count * sizeof (CvPoint2D32f));
	corners2 = (CvPoint2D32f *) cvAlloc (corner_count * sizeof (CvPoint2D32f));
	prev_pyramid = cvCreateImage (cvSize (src_img0->width + 8, src_img0->height / 3), IPL_DEPTH_8U, 1);
	curr_pyramid = cvCreateImage (cvSize (src_img0->width + 8, src_img0->height / 3), IPL_DEPTH_8U, 1);
	status = (char *) cvAlloc (corner_count);
	criteria = cvTermCriteria (CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 64, 0.01);
      
	cvGoodFeaturesToTrack (src_img0, eig_img, temp_img, corners1, &corner_count, 0.001, 5, NULL);
      
	cvCalcOpticalFlowPyrLK (src_img0, src_img1, prev_pyramid, curr_pyramid,
				corners1, corners2, corner_count, cvSize (10, 10), 4, status, NULL, criteria, 0);
      
	for (int i = 0; i < corner_count; i++) {
	  if (status[i]){	
	    // cvLine (dst_img, cvPointFrom32f (corners1[i]), cvPointFrom32f (corners2[i]), CV_RGB (255, 0, 0), 1, CV_AA, 0);
	    (*map_contents).push_back((float)corners1[i].x);
	    (*map_contents).push_back((float)corners1[i].y);
	    (*map_contents).push_back((float)(corners2[i].x - corners1[i].x));
	    (*map_contents).push_back((float)(corners2[i].y - corners1[i].y));
	  }
	}
      
      }
          
      // cv_image = cvCloneImage (dst_img);
    
      // main routine : finish
      // ---------------------------------------------------------------
         
      // CvfdImage* image= new CvfdImage(cv_image);       
      // (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      (*outputMapPtr)[img_it->first] = map_contents;    

      // cvReleaseImage(&cv_image);
      cvReleaseImage(&cv_image0);
      cvReleaseImage(&cv_image1);
      cvReleaseImage(&src_img0);
      cvReleaseImage(&src_img1);
      cvReleaseImage(&dst_img);
      cvReleaseImage(&eig_img);
      cvReleaseImage(&temp_img);
      cvReleaseImage(&prev_pyramid);
      cvReleaseImage(&curr_pyramid);


    }

    cvReleaseImage(&img_in_tmp0);
    cvReleaseImage(&img_in_tmp1);


  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhCalcOpticalFlow)
};
