/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

using namespace std;
using namespace FD;

class CvfdhFeaturePixelSubtraction;
   
DECLARE_NODE(CvfdhFeaturePixelSubtraction)
/*Node
 *
 * @name CvfdhFeaturePixelSubtraction
 * @category FDOPENCV_HARK:Feature
 * @description CvfdhFeaturePixelSubtraction
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 END*/
   
   
class CvfdhFeaturePixelSubtraction : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
      
  typedef struct _PS_FEATURE{
    int I1_x; 
    int I1_y; 
    int I2_x;
    int I2_y;
    int iVal;
  }PixSub_FEATURE;
  
  IplImage* Img;       //入力画像 
  
public:
  CvfdhFeaturePixelSubtraction(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    inOrder = true;
  }

  virtual int translateInput(string inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }

  /*--------------------------------------------------------------------------------*/

  //ピクセル差分特徴量の書き出し
  int OutputPSFeature(int nFeatureNum, PixSub_FEATURE *pdFeatures){
    int  i;
    FILE* fp;

    //出力ファイルのオープン
    if(NULL==(fp=fopen("./output.dat", "a"))){
      printf("Can not open the outfile.\n");
      exit(-1); 
    }

    //ピクセルI1、ピクセルI2、差分値(特徴量)の書き出し
    for(i=0; i<nFeatureNum; i++){
      fprintf(fp, "%d\t%d\t%d\t%d\t%d\n", pdFeatures[i].I1_x, pdFeatures[i].I1_y, pdFeatures[i].I2_x, pdFeatures[i].I2_y, pdFeatures[i].iVal);
    }

    fclose(fp);

    return 0;
  }


  //あらかじめ特徴量の数を算出する
  unsigned long int CountPSFeature(){
    int  i,j;
    unsigned long int fcount;

    //カウント数の初期化
    fcount=0;

    //画像の高さと幅
    for(i=0; i < (Img->width * Img->height); i++){
      for(j=i+1; j < (Img->width * Img->height); j++){
	//特徴量数のカウント
	fcount++;
      }
    }
    cout << " " << Img->width << " " << Img->height << " " << fcount << endl;

    return fcount;
  }


  //ピクセル差分特徴量の抽出
  int PixelSubtraction(PixSub_FEATURE *pdFeatures){
    int  x1,y1;     //I1の座標値
    int  x2,y2;     //I2の座標値
    int  i1_address, i2_address; //I1,I2のIPL構造体での要素番号
    int  fcount;

    int  wStep=Img->widthStep;
    int  bpp=((Img->depth&255)/8)*Img->nChannels; 
    unsigned char *imgSource=(unsigned char*)Img->imageData;

    //カウント数の初期化
    fcount=0;

    //輝度値I1
    for(y1=0; y1<Img->height; y1++){
      for(x1=0; x1<Img->width; x1++){
	i1_address=y1*wStep+x1*bpp;

	//輝度値I2
	for(y2=0; y2<Img->height; y2++){
	  for(x2=0; x2<Img->width; x2++){
	    i2_address=y2*wStep+x2*bpp;

	    //重複回避
	    if(i1_address<i2_address){
	      //特徴量の値と座標を格納
	      pdFeatures[fcount].iVal=imgSource[i1_address]-imgSource[i2_address];
	      pdFeatures[fcount].I1_x=x1;
	      pdFeatures[fcount].I1_y=y1;
	      pdFeatures[fcount].I2_x=x2;
	      pdFeatures[fcount].I2_y=y2;
	      fcount++;
	    }
	  }
	}
      }
    }

    return 0;
  }

  /*--------------------------------------------------------------------------------*/
      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }
    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      Img = cvCloneImage(imagePtr.getImage());
    
      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      // ---------------------------------------------------------------
      // main routine : start

      unsigned long int nPSFeatureNum;

      /*--------特徴抽出--------*/

      //ここに特徴抽出の関数を記述
      nPSFeatureNum=CountPSFeature();

      cout << nPSFeatureNum << endl;
  
      //扱う特徴量の数だけ構造体を確保
      PixSub_FEATURE *pPSFeatures;
      pPSFeatures=new PixSub_FEATURE[nPSFeatureNum];

      //ここに特徴抽出の関数を記述
      PixelSubtraction(pPSFeatures);

      for(int cnt = 0; cnt < nPSFeatureNum; cnt++){
	(*map_contents).push_back((float)pPSFeatures[cnt].iVal);
      }
    
      (*outputMapPtr)[img_it->first] = map_contents;    


      /*------------------------*/

      //特徴量の書き出し
      // OutputPSFeature(nPSFeatureNum, pPSFeatures);

      cout << "done" << endl;

      // End of the transform code

      //開放
      delete []pPSFeatures;

      // IplImage* img_out;
      // img_out = cvCloneImage(Img);

      // main routine : finish
      // ---------------------------------------------------------------
         
      // CvfdImage* image= new CvfdImage(img_out);       
      // (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 
      // cvReleaseImage (&img_out);

    }

    cvReleaseImage(&img_in_tmp);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhFeaturePixelSubtraction)
};
