/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"
#include <vector>

using namespace std;
using namespace FD;

class CvfdhHoughCircle;
   
DECLARE_NODE(CvfdhHoughCircle)
/*Node
 *
 * @name CvfdhHoughCircle
 * @category FDOPENCV_HARK:Segmentation
 * @description CvfdhHoughCircle
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is SrcInfo.
 *
 END*/
   
   
class CvfdhHoughCircle : public BufferedNode {

  int m_imageInID;
  int m_srcOutID;
      
public:
  CvfdhHoughCircle(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcOutID = addOutput("SRCOUT");
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    Vector<ObjectRef>& sources = *new Vector<ObjectRef>;
    (*(outputs[m_srcOutID].buffer))[count] = &sources;

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){
      IplImage* img_in;      
      img_in = cvCloneImage((IplImage*)image_vec[cnt]);

      // ---------------------------------------------------------------
      // main routine : start


      IplImage* gray = cvCreateImage( cvGetSize(img_in), 8, 1 );
      CvMemStorage* storage = cvCreateMemStorage(0);
      cvCvtColor(img_in,  gray, CV_BGR2GRAY );
      cvSmooth( gray, gray, CV_GAUSSIAN, 9, 9 ); // 画像の平滑化，行わないと誤検出が起こりやすい
      CvSeq* circles = cvHoughCircles( gray, storage, CV_HOUGH_GRADIENT, 2, gray->height/4, 200, 100 );
      int i;
      for( i = 0; i < circles->total; i++ ){
	float* p = (float*)cvGetSeqElem( circles, i );
	//cvCircle( img_in, cvPoint(cvRound(p[0]), cvRound(p[1])), 3, CV_RGB(0,255,0), -1, 8, 0 );
	//cvCircle( img_in, cvPoint(cvRound(p[0]), cvRound(p[1])), cvRound(p[2]), CV_RGB(255,0,0), 3, 8, 0 );

	RCPtr<SrcInfo> src(new SrcInfo);
	src->roi_c[0] = cvRound(p[0]);
	src->roi_c[1] = cvRound(p[1]);
	src->roi_xy[0] = 2 * cvRound(p[2]);
	src->roi_xy[1] = 2 * cvRound(p[2]);
	src->img_size[0] = img_in->width;
	src->img_size[1] = img_in->height;
      
	sources.push_back(src);

      }

      cvReleaseImage (&img_in);
      cvReleaseImage (&gray);

    }

    cvReleaseImage(&img_in_tmp);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhHoughCircle)
};
