/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

using namespace std;
using namespace FD;

class CvfdhImageMoment;
   
DECLARE_NODE(CvfdhImageMoment)
/*Node
 *
 * @name CvfdhImageMoment
 * @category FDOPENCV_HARK:Feature
 * @description CvfdhImageMoment
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 END*/
   
   
class CvfdhImageMoment : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
      
public:
  CvfdhImageMoment(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    inOrder = true;
  }      

  virtual int translateInput(string inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }
    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      IplImage* img_in;      
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      img_in = cvCloneImage(imagePtr.getImage());
    
      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      // ---------------------------------------------------------------
      // main routine : start

      char text[10][30];
      int i;
      double spatial_moment, central_moment, norm_c_moment;

      CvFont font;
      CvSize text_size;
      CvMoments moments;
      CvHuMoments hu_moments;

      if (img_in->nChannels == 3 && cvGetImageCOI (img_in) == 0)
	cvSetImageCOI (img_in, 1);

      // 入力画像の３次までの画像モーメントを計算する
      cvMoments (img_in, &moments, 0);
      cvSetImageCOI (img_in, 0);

      // モーメントやHuモーメント不変量を，得られたCvMoments構造体の値を使って計算する．
      spatial_moment = cvGetSpatialMoment (&moments, 0, 0);
      central_moment = cvGetCentralMoment (&moments, 0, 0);
      norm_c_moment = cvGetNormalizedCentralMoment (&moments, 0, 0);
      cvGetHuMoments (&moments, &hu_moments);

      // 得られたモーメントやHuモーメント不変量を文字として画像に描画
      // cvInitFont (&font, CV_FONT_HERSHEY_SIMPLEX, 1.0, 1.0, 0, 2, 8);
      // snprintf (text[0], 30, "spatial=%.3f", spatial_moment);
      // snprintf (text[1], 30, "central=%.3f", central_moment);
      // snprintf (text[2], 30, "norm=%.3f", spatial_moment);
      // snprintf (text[3], 30, "hu1=%f", hu_moments.hu1);
      // snprintf (text[4], 30, "hu2=%f", hu_moments.hu2);
      // snprintf (text[5], 30, "hu3=%f", hu_moments.hu3);
      // snprintf (text[6], 30, "hu4=%f", hu_moments.hu4);
      // snprintf (text[7], 30, "hu5=%f", hu_moments.hu5);
      // snprintf (text[8], 30, "hu6=%f", hu_moments.hu6);
      // snprintf (text[9], 30, "hu7=%f", hu_moments.hu7);
      // cvGetTextSize (text[0], &font, &text_size, 0);
      // for (i = 0; i < 10; i++)
      //   cvPutText (img_in, text[i], cvPoint (10, (text_size.height + 3) * (i + 1)), &font, cvScalarAll (0));

      (*map_contents).push_back((float)spatial_moment);
      (*map_contents).push_back((float)central_moment);
      (*map_contents).push_back((float)norm_c_moment);
      (*map_contents).push_back((float)hu_moments.hu1);
      (*map_contents).push_back((float)hu_moments.hu2);
      (*map_contents).push_back((float)hu_moments.hu3);
      (*map_contents).push_back((float)hu_moments.hu4);
      (*map_contents).push_back((float)hu_moments.hu5);
      (*map_contents).push_back((float)hu_moments.hu6);
      (*map_contents).push_back((float)hu_moments.hu7);

      (*outputMapPtr)[img_it->first] = map_contents;    

      // IplImage* img_out;
      // img_out = cvCloneImage(img_in);

      // main routine : finish
      // ---------------------------------------------------------------
         
      // CvfdImage* image= new CvfdImage(img_out);       
      // (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage (&img_in);
      // cvReleaseImage (&img_out);

    }

    cvReleaseImage(&img_in_tmp);



  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhImageMoment)
};
