/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"
#include <vector>

#if ENABLE_OPENCV_2_0_0

#include "cvaux.h"
#include "highgui.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>

using namespace std;
using namespace FD;
// using namespace cv;

class CvfdhPeopleDetect;
   
DECLARE_NODE(CvfdhPeopleDetect)
/*Node
 *
 * @name CvfdhPeopleDetect
 * @category FDOPENCV_HARK:Segmentation
 * @description CvfdhPeopleDetect
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @parameter_name HIT_THRESHOLD
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_description Threshold for the distance between features and SVM classifying plane. See gpu::HOGDescriptor::detect() for details.
 *
 * @parameter_name WIN_STRIDE
 * @parameter_type object
 * @parameter_value <Vector<int> 8 8>
 * @parameter_description Window stride. It must be a multiple of block stride.
 *
 * @parameter_name PADDING
 * @parameter_type object
 * @parameter_value <Vector<int> 32 32>
 * @parameter_description Mock parameter to keep the CPU interface compatibility. It must be (0,0).
 *
 * @parameter_name SCALE0
 * @parameter_type float
 * @parameter_value 1.05
 * @parameter_description Coefficient of the detection window increase.
 *
 * @parameter_name GROUP_THRESHOLD
 * @parameter_type int
 * @parameter_value 2
 * @parameter_description Coefficient to regulate the similarity threshold. When detected, some objects can be covered by many rectangles. 0 means not to perform grouping. See groupRectangles().
 *
 END*/
   
   
class CvfdhPeopleDetect : public BufferedNode {

  int m_imageInID;
  int m_srcOutID;
  float hit_threshold;
  Vector<int> win_stride;  
  Vector<int> padding;  
  float scale0;
  int group_threshold;
      
public:
  CvfdhPeopleDetect(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcOutID = addOutput("SRCOUT");

    hit_threshold = dereference_cast<float>(parameters.get("HIT_THRESHOLD"));

    win_stride = object_cast<Vector<int> >(parameters.get("WIN_STRIDE"));    
    if (win_stride.size() != 2) {
      throw new NodeException(this, string("WIN_STRIDE is not correctly set."), __FILE__, __LINE__);
    }

    padding = object_cast<Vector<int> >(parameters.get("PADDING"));    
    if (padding.size() != 2) {
      throw new NodeException(this, string("PADDING is not correctly set."), __FILE__, __LINE__);
    }

    scale0 = dereference_cast<float>(parameters.get("SCALE0"));
    group_threshold = dereference_cast<int>(parameters.get("GROUP_THRESHOLD"));

    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    Vector<ObjectRef>& sources = *new Vector<ObjectRef>;
    (*(outputs[m_srcOutID].buffer))[count] = &sources;

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){
      IplImage* cv_image;      
      cv_image = cvCloneImage((IplImage*)image_vec[cnt]);
    
      // ---------------------------------------------------------------
      // main routine : start

      cv::Mat img;
      img = cv_image;
    
      cv::HOGDescriptor hog;
      hog.setSVMDetector(cv::HOGDescriptor::getDefaultPeopleDetector());

      fflush(stdout);
      vector<cv::Rect> found, found_filtered;
      double t = (double)cv::getTickCount();
      // run the detector with default parameters. to get a higher hit-rate
      // (and more false alarms, respectively), decrease the hitThreshold and
      // groupThreshold (set groupThreshold to 0 to turn off the grouping completely).
      int can = img.channels();
      hog.detectMultiScale(img, found, hit_threshold, cv::Size(win_stride[0],win_stride[1]), cv::Size(padding[0],padding[1]), scale0, group_threshold);
      t = (double)cv::getTickCount() - t;
      printf("tdetection time = %gms\n", t*1000./cv::getTickFrequency());
      size_t i, j;
      for( i = 0; i < found.size(); i++ )
	{
	  cv::Rect r = found[i];
	  for( j = 0; j < found.size(); j++ )
	    if( j != i && (r & found[j]) == r)
	      break;
	  if( j == found.size() )
	    found_filtered.push_back(r);
	}
      for( i = 0; i < found_filtered.size(); i++ )
	{
	  cv::Rect r = found_filtered[i];
	  // the HOG detector returns slightly larger rectangles than the real objects.
	  // so we slightly shrink the rectangles to get a nicer output.
	  r.x += cvRound(r.width*0.1);
	  r.width = cvRound(r.width*0.8);
	  r.y += cvRound(r.height*0.07);
	  r.height = cvRound(r.height*0.8);
	  //rectangle(img, r.tl(), r.br(), cv::Scalar(0,255,0), 3);

	  RCPtr<SrcInfo> src(new SrcInfo);
	  src->roi_c[0] = (int)((r.tl().x + r.br().x) / 2);
	  src->roi_c[1] = (int)((r.tl().y + r.br().y) / 2);
	  src->roi_xy[0] = (int)abs(r.tl().x - r.br().x);
	  src->roi_xy[1] = (int)abs(r.tl().y - r.br().y);
	  src->img_size[0] = cv_image->width;
	  src->img_size[1] = cv_image->height;

	  sources.push_back(src);
	
	}

      cvReleaseImage(&cv_image);
    }

    cvReleaseImage(&img_in_tmp);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhPeopleDetect)
};

#endif
