/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhPyrMeanShiftFiltering;
   
DECLARE_NODE(CvfdhPyrMeanShiftFiltering)
/*Node
 *
 * @name CvfdhPyrMeanShiftFiltering
 * @category FDOPENCV_HARK:Filter
 * @description CvfdhPyrMeanShiftFiltering
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @parameter_name LEVEL
 * @parameter_type int
 * @parameter_value 2
 * @parameter_description 
 *
 END*/
   
   
class CvfdhPyrMeanShiftFiltering : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  int level;
      
public:
  CvfdhPyrMeanShiftFiltering(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    level = dereference_cast<int>(parameters.get("LEVEL"));
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {

      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }

      IplImage* cv_image;
      cv_image = cvCloneImage(imagePtr->getImage());
    
      IplImage *dst_img;
      CvRect roi;

      // 領域分割のためにROIをセットする
      roi.x = roi.y = 0;
      roi.width = cv_image->width & -(1 << level);
      roi.height = cv_image->height & -(1 << level);
      cvSetImageROI (cv_image, roi);

      // 分割結果画像出力用の画像領域を確保し，領域分割を実行
      dst_img = cvCloneImage (cv_image);
    
      cvPyrMeanShiftFiltering (cv_image, dst_img, 30.0, 30.0, level,
			       cvTermCriteria (CV_TERMCRIT_ITER + CV_TERMCRIT_EPS, 5, 1));

      IplImage* img_out;
      img_out = cvCloneImage(dst_img);

      CvfdImage* image= new CvfdImage(img_out);  
         
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage(&cv_image);
      cvReleaseImage(&dst_img);
      cvReleaseImage(&img_out);

      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* cv_image;
	cv_image = cvCloneImage(imagePtr.getImage());

	IplImage *dst_img;
	CvRect roi;
	
	// 領域分割のためにROIをセットする
	roi.x = roi.y = 0;
	roi.width = cv_image->width & -(1 << level);
	roi.height = cv_image->height & -(1 << level);
	cvSetImageROI (cv_image, roi);

	// 分割結果画像出力用の画像領域を確保し，領域分割を実行
	dst_img = cvCloneImage (cv_image);
	
	cvPyrMeanShiftFiltering (cv_image, dst_img, 30.0, 30.0, level,
				 cvTermCriteria (CV_TERMCRIT_ITER + CV_TERMCRIT_EPS, 5, 1));
	
	IplImage* img_out;
	img_out = cvCloneImage(dst_img);
	
	CvfdImage* image= new CvfdImage(img_out);  

	(*outputMapPtr)[it->first] = image;

	cvReleaseImage(&cv_image);
	cvReleaseImage(&dst_img);
	cvReleaseImage(&img_out);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhPyrMeanShiftFiltering)
};
