/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"

using namespace std;
using namespace FD;

class CvfdSmooth;
   
DECLARE_NODE(CvfdSmooth)
/*Node
 *
 * @name CvfdSmooth
 * @category FDOPENCV_ORG:Morphological Operations
 * @description Smooths the image
 *
 * @input_name IMAGEIN
 * @input_description Input image
 * @input_type CvfdImage
 * 
 * @output_name IMAGEOUT
 * @output_description Output image
 * @output_type CvfdImage
 *
 * @parameter_name SMOOTH_TYPE
 * @parameter_type string
 * @parameter_value CV_BLUR_NO_SCALE;CV_BLUR;CV_GAUSSIAN;CV_MEDIAN;CV_BILATERAL
 * @parameter_description Type of the smoothing
 * CV_BLUR_NO_SCALE (simple blur with no scaling) - summation over a pixel param1恊aram2 neighborhood. If the neighborhood size may vary, one may precompute integral image with cvIntegral function.
 * CV_BLUR (simple blur) - summation over a pixel param1xparam2 neighborhood with subsequent scaling by 1/(param1*param2).
 * CV_GAUSSIAN (gaussian blur) - convolving image with param1ラparam2 Gaussian kernel.
 * CV_MEDIAN (median blur) - finding median of param1ラparam1 neighborhood (i.e. the neighborhood is square).
 * CV_BILATERAL (bilateral filter) - applying bilateral 3x3 filtering with color sigma=param1 and space sigma=param2.    
 *
 * @parameter_name PARAM1
 * @parameter_type int
 * @parameter_value 3
 * @parameter_description The first parameter of smoothing operation
 *
 * @parameter_name PARAM2
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The second parameter of smoothing operation. In case of simple scaled/non-scaled and Gaussian blur if param2 is zero, it is set to param1
 *
 * @parameter_name PARAM3
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description In case of Gaussian parameter this parameter may specify Gaussian sigma (standard deviation). If it is zero, it is calculated from the kernel size
 END*/
   
   
class CvfdSmooth : public BufferedNode {
  //Input ID
  int m_imageInID;
  //Output ID
  int m_imageOutID;
  //Parameters
  string m_smoothType;
  int m_param1;
  int m_param2;
  int m_param3;
  //map of the smooth_type
  map<string,int> m_smoothTypeMap;
      
public:
  CvfdSmooth(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_imageInID = addInput("IMAGEIN");
    //add outputs
    m_imageOutID = addOutput("IMAGEOUT");
    //Initialize parameters
    m_param1 = dereference_cast<int>(parameters.get("PARAM1"));
    m_param2 = dereference_cast<int>(parameters.get("PARAM2"));
    m_param3 = dereference_cast<int>(parameters.get("PARAM3"));
    m_smoothType = object_cast<String>(parameters.get("SMOOTH_TYPE"));         
         
    //Initialize the smoothtype map
    m_smoothTypeMap["CV_BLUR_NO_SCALE"] = CV_BLUR_NO_SCALE;
    m_smoothTypeMap["CV_BLUR"] = CV_BLUR;
    m_smoothTypeMap["CV_GAUSSIAN"] = CV_GAUSSIAN; 
    m_smoothTypeMap["CV_MEDIAN"] = CV_MEDIAN;
    m_smoothTypeMap["CV_BILATERAL"] = CV_BILATERAL;         
         
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {
    //Read the inputs
    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);  
         
    //Handle
    CvfdImage* image = new CvfdImage(imagePtr->getImage());
         
    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent );
    __BEGIN__; 
    OPENCV_CALL( cvSmooth( imagePtr->getImage(), image->getImage(),
			   m_smoothTypeMap[m_smoothType],
			   m_param1, m_param2, m_param3 ));         
    __END__;
    cvSetErrMode( status );
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to smooth the image: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      }         
    out[count] = ObjectRef(image);
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdSmooth)
};
